@@ -26,6 +26,7 @@ local lua_redis = require "lua_redis"
 local fun = require "fun"
 local lua_maps = require "lua_maps"
 local lua_util = require "lua_util"
+local lua_verdict = require "lua_verdict"
 local rspamd_hash = require "rspamd_cryptobox_hash"
 local lua_selectors = require "lua_selectors"
 local ts = require("tableshape").types
@@ -50,6 +51,7 @@ local settings = {
   expire = 60 * 60 * 24 * 2, -- 2 days by default
   limits = {},
   allow_local = false,
+  prefilter = true,
 }
 
 -- Checks bucket, updating it if needed
@@ -666,7 +668,7 @@ local function ratelimit_update_cb(task)
       return
     end
 
-    local verdict = lua_util.get_task_verdict(task)
+    local verdict = lua_verdict.get_specific_verdict(N, task)
     local _,nrcpt = task:has_recipients('smtp')
     if not nrcpt or nrcpt <= 0 then
       nrcpt = 1
@@ -855,20 +857,32 @@ if opts then
     lua_util.disable_module(N, "redis")
   else
     local s = {
-      type = 'prefilter,nostat',
+      type = settings.prefilter and 'prefilter' or 'callback',
       name = 'RATELIMIT_CHECK',
       priority = 7,
       callback = ratelimit_cb,
-      flags = 'empty',
+      flags = 'empty,nostat',
     }
 
+    local id = rspamd_config:register_symbol(s)
+
+    if settings.info_symbol then
+      rspamd_config:register_symbol{
+        type = 'virtual',
+        name = settings.info_symbol,
+        score = 0.0,
+        parent = id
+      }
+    end
     if settings.symbol then
-      s.name = settings.symbol
-    elseif settings.info_symbol then
-      s.name = settings.info_symbol
+      rspamd_config:register_symbol{
+        type = 'virtual',
+        name = settings.symbol,
+        score = 0.0, -- Might be overridden if needed
+        parent = id
+      }
     end
 
-    rspamd_config:register_symbol(s)
     rspamd_config:register_symbol {
       type = 'idempotent',
       name = 'RATELIMIT_UPDATE',
