@@ -26,6 +26,8 @@ local E = {}
 local rcvd_cb_id = rspamd_config:register_symbol{
   name = 'CHECK_RECEIVED',
   type = 'callback',
+  score = 0.0,
+  group = 'headers',
   callback = function(task)
     local cnts = {
       [1] = 'ONE',
@@ -58,7 +60,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = 'No received',
+  description = 'Message has no Received headers',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -66,7 +68,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = 'One received',
+  description = 'Message has one Received header',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -74,7 +76,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = 'Two received',
+  description = 'Message has two Received headers',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -82,7 +84,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = '3-5 received',
+  description = 'Message has 3-5 Received headers',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -90,7 +92,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = '5-7 received',
+  description = 'Message has 5-7 Received headers',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -98,7 +100,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = '7-11 received',
+  description = 'Message has 7-11 Received headers',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -106,13 +108,16 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = rcvd_cb_id,
   type = 'virtual',
-  description = '12+ received',
+  description = 'Message has 12 or more Received headers',
   group = 'headers',
 }
 
 local prio_cb_id = rspamd_config:register_symbol {
   name = 'HAS_X_PRIO',
   type = 'callback',
+  description = 'X-Priority check callback rule',
+  score = 0.0,
+  group = 'headers',
   callback = function (task)
      local cnts = {
       [1] = 'ONE',
@@ -140,7 +145,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = prio_cb_id,
   type = 'virtual',
-  description = 'Priority 0',
+  description = 'Message has X-Priority header set to 0',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -148,7 +153,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = prio_cb_id,
   type = 'virtual',
-  description = 'Priority 1',
+  description = 'Message has X-Priority header set to 1',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -156,7 +161,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = prio_cb_id,
   type = 'virtual',
-  description = 'Priority 2',
+  description = 'Message has X-Priority header set to 2',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -164,7 +169,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = prio_cb_id,
   type = 'virtual',
-  description = 'Priority 3-4',
+  description = 'Message has X-Priority header set to 3 or 4',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -172,7 +177,7 @@ rspamd_config:register_symbol{
   score = 0.0,
   parent = prio_cb_id,
   type = 'virtual',
-  description = 'Priority 5+',
+  description = 'Message has X-Priority header set to 5 or higher',
   group = 'headers',
 }
 
@@ -180,10 +185,16 @@ local function get_raw_header(task, name)
   return ((task:get_header_full(name) or {})[1] or {})['value']
 end
 
-local check_replyto_id = rspamd_config:register_callback_symbol('CHECK_REPLYTO', 1.0,
-  function (task)
+local check_replyto_id = rspamd_config:register_symbol({
+  type = 'callback',
+  name = 'CHECK_REPLYTO',
+  score = 0.0,
+  group = 'headers',
+  callback = function(task)
     local replyto = get_raw_header(task, 'Reply-To')
-    if not replyto then return false end
+    if not replyto then
+      return false
+    end
     local rt = util.parse_mail_address(replyto, task:get_mempool())
     if not (rt and rt[1] and (string.len(rt[1].addr) > 0)) then
       task:insert_result('REPLYTO_UNPARSEABLE', 1.0)
@@ -203,9 +214,11 @@ local check_replyto_id = rspamd_config:register_callback_symbol('CHECK_REPLYTO',
     end
 
     -- See if Reply-To matches From in some way
-    local from = task:get_from(2)
+    local from = task:get_from{'mime', 'orig'}
     local from_h = get_raw_header(task, 'From')
-    if not (from and from[1]) then return false end
+    if not (from and from[1]) then
+      return false
+    end
     if (from_h and from_h == replyto) then
       -- From and Reply-To are identical
       task:insert_result('REPLYTO_EQ_FROM', 1.0)
@@ -221,17 +234,17 @@ local check_replyto_id = rspamd_config:register_callback_symbol('CHECK_REPLYTO',
             -- See if Reply-To matches the To address
             local to = task:get_recipients(2)
             if (to and to[1] and to[1].addr:lower() == rt[1].addr:lower()) then
-                -- Ignore this for mailing-lists and automatic submissions
-                if (not (task:get_header('List-Unsubscribe') or
-                         task:get_header('X-To-Get-Off-This-List') or
-                         task:get_header('X-List') or
-                         task:get_header('Auto-Submitted')))
-                then
-                   task:insert_result('REPLYTO_EQ_TO_ADDR', 1.0)
-                end
+              -- Ignore this for mailing-lists and automatic submissions
+              if (not (task:get_header('List-Unsubscribe') or
+                  task:get_header('X-To-Get-Off-This-List') or
+                  task:get_header('X-List') or
+                  task:get_header('Auto-Submitted')))
+              then
+                task:insert_result('REPLYTO_EQ_TO_ADDR', 1.0)
+              end
             else
               task:insert_result('REPLYTO_DOM_NEQ_FROM_DOM', 1.0)
-	    end
+            end
           end
         end
         -- See if the Display Names match
@@ -242,7 +255,7 @@ local check_replyto_id = rspamd_config:register_callback_symbol('CHECK_REPLYTO',
       end
     end
   end
-)
+})
 
 rspamd_config:register_symbol{
   name = 'REPLYTO_UNPARSEABLE',
@@ -322,48 +335,43 @@ rspamd_config:register_dependency('CHECK_REPLYTO', 'CHECK_FROM')
 local check_mime_id = rspamd_config:register_symbol{
   name = 'CHECK_MIME',
   type = 'callback',
+  group = 'headers',
+  score = 0.0,
   callback = function(task)
-    local parts = task:get_parts()
-    if not parts then return false end
-
-    -- Make sure there is a MIME-Version header
-    local mv = task:get_header('MIME-Version')
+    -- Check if there is a MIME-Version header
     local missing_mime = false
-    if (not mv) then
+    if not task:has_header('MIME-Version') then
       missing_mime = true
     end
 
+    -- Check presense of MIME specific headers
+    local has_ct_header = task:has_header('Content-Type')
+    local has_cte_header = task:has_header('Content-Transfer-Encoding')
+
+    -- Add the symbol if we have MIME headers, but no MIME-Version
+    -- (do not add the symbol for RFC822 messages)
+    if (has_ct_header or has_cte_header) and missing_mime then
+      task:insert_result('MISSING_MIME_VERSION', 1.0)
+    end
+
     local found_ma = false
     local found_plain = false
     local found_html = false
-    local cte_7bit = false
 
-    for _,p in ipairs(parts) do
-      local mtype,subtype = p:get_type()
+    for _, p in ipairs(task:get_parts()) do
+      local mtype, subtype = p:get_type()
       local ctype = mtype:lower() .. '/' .. subtype:lower()
       if (ctype == 'multipart/alternative') then
         found_ma = true
       end
       if (ctype == 'text/plain') then
-        if p:get_cte() == '7bit' then
-          cte_7bit = true
-        end
         found_plain = true
       end
       if (ctype == 'text/html') then
-        if p:get_cte() == '7bit' then
-          cte_7bit = true
-        end
         found_html = true
       end
     end
 
-    if missing_mime then
-      if not (not found_ma and ((found_plain or found_html) and cte_7bit)) then
-        task:insert_result('MISSING_MIME_VERSION', 1.0)
-      end
-    end
-
     if (found_ma) then
       if (not found_plain) then
         task:insert_result('MIME_MA_MISSING_TEXT', 1.0)
@@ -380,7 +388,7 @@ rspamd_config:register_symbol{
   score = 2.0,
   parent = check_mime_id,
   type = 'virtual',
-  description = 'MIME-Version header is missing',
+  description = 'MIME-Version header is missing in MIME message',
   group = 'headers',
 }
 rspamd_config:register_symbol{
@@ -420,6 +428,7 @@ rspamd_config.PREVIOUSLY_DELIVERED = {
     end
   end,
   description = 'Message either to a list or was forwarded',
+  group = 'headers',
   score = 0.0
 }
 rspamd_config.BROKEN_HEADERS = {
@@ -567,15 +576,8 @@ rspamd_config.MISSING_FROM = {
 rspamd_config.MULTIPLE_FROM = {
   callback = function(task)
     local from = task:get_from('mime')
-    if from and from[1] then
-      if #from > 1 then
-        return true,1.0,table.concat(
-            fun.totable(
-                fun.map(function(a) return a.addr end,
-                    fun.filter(function(a) return a.addr and a.addr ~= '' end,
-                        from))),
-            ',')
-      end
+    if from and from[2] then
+      return true, 1.0, fun.totable(fun.map(function(a) return a.raw end, from))
     end
     return false
   end,
@@ -586,64 +588,59 @@ rspamd_config.MULTIPLE_FROM = {
 
 rspamd_config.MV_CASE = {
   callback = function (task)
-    local mv = task:get_header('Mime-Version', true)
-    if (mv) then return true end
+    return task:has_header('Mime-Version', true)
   end,
   description = 'Mime-Version .vs. MIME-Version',
   score = 0.5,
-  group = 'headers',
-}
-
-rspamd_config.FAKE_REPLY = {
-  callback = function (task)
-    local subject = task:get_header('Subject')
-    if (subject and subject:lower():find('^re:')) then
-      local ref = task:get_header('References')
-      local rt  = task:get_header('In-Reply-To')
-      if (not (ref or rt)) then return true end
-    end
-    return false
-  end,
-  description = 'Fake reply',
-  score = 1.0,
   group = 'headers'
 }
 
 local check_from_id = rspamd_config:register_symbol{
   name = 'CHECK_FROM',
   type = 'callback',
+  score = 0.0,
+  group = 'headers',
   callback = function(task)
     local envfrom = task:get_from(1)
     local from = task:get_from(2)
-    if (from and from[1] and (from[1].name == nil or from[1].name == '' )) then
-      task:insert_result('FROM_NO_DN', 1.0)
-    elseif (from and from[1] and from[1].name and
-            util.strequal_caseless(from[1].name, from[1].addr)) then
-      task:insert_result('FROM_DN_EQ_ADDR', 1.0)
-    elseif (from and from[1] and from[1].name and from[1].name ~= '') then
-      task:insert_result('FROM_HAS_DN', 1.0)
-      -- Look for Mr/Mrs/Dr titles
-      local n = from[1].name:lower()
-      local match, match_end
-      match, match_end = n:find('^mrs?[%.%s]')
-      if match then
-        task:insert_result('FROM_NAME_HAS_TITLE', 1.0, n:sub(match, match_end-1))
-      end
-      match, match_end = n:find('^dr[%.%s]')
-      if match then
-        task:insert_result('FROM_NAME_HAS_TITLE', 1.0, n:sub(match, match_end-1))
-      end
-      -- Check for excess spaces
-      if n:find('%s%s') then
-        task:insert_result('FROM_NAME_EXCESS_SPACE', 1.0)
-      end
+    if (envfrom and envfrom[1] and not envfrom[1]["flags"]["valid"]) then
+      task:insert_result('ENVFROM_INVALID', 1.0)
     end
-    if (envfrom and from and envfrom[1] and from[1] and
-        util.strequal_caseless(envfrom[1].addr, from[1].addr))
-    then
-      task:insert_result('FROM_EQ_ENVFROM', 1.0)
-    elseif (envfrom and envfrom[1] and envfrom[1].addr) then
-      task:insert_result('FROM_NEQ_ENVFROM', 1.0, ((from or E)[1] or E).addr or '', envfrom[1].addr)
+    if (from and from[1]) then
+      if not (from[1]["flags"]["valid"]) then
+        task:insert_result('FROM_INVALID', 1.0)
+      end
+      if (from[1].name == nil or from[1].name == '' ) then
+        task:insert_result('FROM_NO_DN', 1.0)
+      elseif (from[1].name and
+            util.strequal_caseless(from[1].name, from[1].addr)) then
+        task:insert_result('FROM_DN_EQ_ADDR', 1.0)
+      elseif (from[1].name and from[1].name ~= '') then
+        task:insert_result('FROM_HAS_DN', 1.0)
+        -- Look for Mr/Mrs/Dr titles
+        local n = from[1].name:lower()
+        local match, match_end
+        match, match_end = n:find('^mrs?[%.%s]')
+        if match then
+          task:insert_result('FROM_NAME_HAS_TITLE', 1.0, n:sub(match, match_end-1))
+        end
+        match, match_end = n:find('^dr[%.%s]')
+        if match then
+          task:insert_result('FROM_NAME_HAS_TITLE', 1.0, n:sub(match, match_end-1))
+        end
+        -- Check for excess spaces
+        if n:find('%s%s') then
+          task:insert_result('FROM_NAME_EXCESS_SPACE', 1.0)
+        end
+      end
+
+      if envfrom then
+        if util.strequal_caseless(envfrom[1].addr, from[1].addr) then
+          task:insert_result('FROM_EQ_ENVFROM', 1.0)
+        elseif envfrom[1].addr ~= '' then
+          task:insert_result('FROM_NEQ_ENVFROM', 1.0, from[1].addr, envfrom[1].addr)
+        end
+      end
     end
 
     local to = task:get_recipients(2)
@@ -659,9 +656,25 @@ local check_from_id = rspamd_config:register_symbol{
   end
 }
 
+rspamd_config:register_symbol{
+  name = 'ENVFROM_INVALID',
+  score = 2.0,
+  group = 'headers',
+  parent = check_from_id,
+  type = 'virtual',
+  description = 'Envelope from does not have a valid format',
+}
+rspamd_config:register_symbol{
+  name = 'FROM_INVALID',
+  score = 2.0,
+  group = 'headers',
+  parent = check_from_id,
+  type = 'virtual',
+  description = 'From header does not have a valid format',
+}
 rspamd_config:register_symbol{
   name = 'FROM_NO_DN',
-  score = 0,
+  score = 0.0,
   group = 'headers',
   parent = check_from_id,
   type = 'virtual',
@@ -735,6 +748,8 @@ rspamd_config:register_symbol{
 local check_to_cc_id = rspamd_config:register_symbol{
   name = 'CHECK_TO_CC',
   type = 'callback',
+  score = 0.0,
+  group = 'headers,mime',
   callback = function(task)
     local rcpts = task:get_recipients(1)
     local to = task:get_recipients(2)
@@ -935,6 +950,7 @@ rspamd_config:register_symbol{
   description = 'Some of the recipients match the envelope',
 }
 
+-- TODO: rewrite this rule, it should not touch headers directly
 rspamd_config.CTYPE_MISSING_DISPOSITION = {
   callback = function(task)
     local parts = task:get_parts()
@@ -949,6 +965,18 @@ rspamd_config.CTYPE_MISSING_DISPOSITION = {
           then
             return false
           end
+
+          local parent = p:get_parent()
+
+          if parent then
+            local t,st = parent:get_type()
+
+            if t == 'multipart' and st == 'encrypted' then
+              -- Special case
+              return false
+            end
+          end
+
           return true
         end
       end
@@ -957,7 +985,7 @@ rspamd_config.CTYPE_MISSING_DISPOSITION = {
   end,
   description = 'Binary content-type not specified as an attachment',
   score = 4.0,
-  group = 'headers'
+  group = 'mime'
 }
 
 rspamd_config.CTYPE_MIXED_BOGUS = {
@@ -969,14 +997,19 @@ rspamd_config.CTYPE_MIXED_BOGUS = {
     if (not ct:lower():match('^multipart/mixed')) then return false end
     local found = false
     -- Check each part and look for a part that isn't multipart/* or text/plain or text/html
+    local ntext_parts = 0
     for _,p in ipairs(parts) do
-      local pct = p:get_header('Content-Type')
-      if (pct) then
-        pct = pct:lower()
-        if not ((pct:match('^multipart/') or
-            pct:match('^text/plain') or
-            pct:match('^text/html'))) then
+      local mtype,_ = p:get_type()
+      if mtype then
+        if mtype == 'text' and not p:is_attachment() then
+          ntext_parts = ntext_parts + 1
+          if ntext_parts > 2 then
+            found = true
+            break
+          end
+        elseif mtype ~= 'multipart' then
           found = true
+          break
         end
       end
     end
@@ -985,7 +1018,7 @@ rspamd_config.CTYPE_MIXED_BOGUS = {
   end,
   description = 'multipart/mixed without non-textual part',
   score = 1.0,
-  group = 'headers'
+  group = 'mime'
 }
 
 local function check_for_base64_text(part)
@@ -1021,7 +1054,26 @@ rspamd_config.MIME_BASE64_TEXT = {
   end,
   description = 'Has text part encoded in base64',
   score = 0.1,
-  group = 'headers'
+  group = 'mime'
+}
+
+rspamd_config.MIME_BASE64_TEXT_BOGUS = {
+  callback = function(task)
+    local parts = task:get_text_parts()
+    if (not parts) then return false end
+    -- Check each part and look for base64 encoded text parts
+    -- where the part does not have any 8bit characters within it
+    for _, part in ipairs(parts) do
+      local mimepart = part:get_mimepart();
+      if (check_for_base64_text(mimepart) and not part:has_8bit()) then
+        return true
+      end
+    end
+    return false
+  end,
+  description = 'Has text part encoded in base64 that does not contain any 8bit characters',
+  score = 1.0,
+  group = 'mime'
 }
 
 local function is_8bit_addr(addr)
@@ -1062,10 +1114,9 @@ rspamd_config.INVALID_RCPT_8BIT = {
 
 rspamd_config.XM_CASE = {
   callback = function (task)
-    local xm = task:get_header('X-mailer', true)
-    if (xm) then return true end
+    return task:has_header('X-mailer', true)
   end,
   description = 'X-mailer .vs. X-Mailer',
   score = 0.5,
-  group = 'headers',
+  group = 'headers'
 }
